/*
 *  WMBlueCPU - a cpu monitor
 *
 *  Copyright (C) 2003 Draghicioiu Mihai Andrei <misuceldestept@go.ro>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>

#include "cpu.h"
#include "mwm.h"

#include "pixmap.xpm"
#include "icon.xpm"

#define VERSION "0.6"
#define WINDOW_WIDTH 64
#define WINDOW_HEIGHT 64
#define WINDOW_NAME "WMBlueCPU"
#define ICON_NAME "wmbluecpu"
#define CLASS_NAME "wmbluecpu"
#define CLASS_CLASS "WMBlueCPU"
#define GRAPH_HEIGHT 12
#define GRAPH_LENGTH 18
#define GRAPH_X 6
#define GRAPH_Y 24
#define GRAPH_PT_W 3
#define GRAPH_PT_H 3
#define PERCENT_X 28
#define PERCENT_Y 16
#define CPUNUM_X 13
#define CPUNUM_Y 16

#define OPT_DISPLAY NULL
#define OPT_MILISECS 1000
#define OPT_BGCOLOR  "rgb:16/7A/EA"
#define OPT_ONCOLOR  "rgb:77/CD/FF"
#define OPT_OFFCOLOR "rgb:00/9F/FF"
#define OPT_WINDOW 0
#define OPT_SHAPE 1
#define OPT_CPUNUM 0

int            argc;
char         **argv;
Display       *display;
int            screen;
Colormap       colormap;
Visual        *visual;
int            depth;
GC             gc;
Window         rootwindow, window, iconwindow, mapwindow;
XEvent         event;
int            exitloop;
Atom           wm_delete_window;
Atom           _motif_wm_hints;
MWMHints       mwmhints;
Pixmap         buffer, pixmap, pixmask, iconpixmap;
unsigned long  bgcolor, offcolor, oncolor;

char *opt_display  = OPT_DISPLAY;
int   opt_milisecs = OPT_MILISECS;
char *opt_bgcolor  = OPT_BGCOLOR;
char *opt_oncolor  = OPT_ONCOLOR;
char *opt_offcolor = OPT_OFFCOLOR;
int   opt_window   = OPT_WINDOW;
int   opt_shape    = OPT_SHAPE;
int   opt_cpunum   = OPT_CPUNUM;

unsigned long get_color(char *colorname)
{
 XColor color, color2;
 color.pixel = 0;
 XAllocNamedColor(display, colormap, colorname, &color2, &color);
 return color2.pixel;
}

void bad_option(int arg)
{
 fprintf(stderr, "%s needs an argument.\n"
                 "Try `%s --help' for more information.\n",
                 argv[arg-1], argv[0]);
 exit(1);
}

void print_usage()
{
 printf("Usage: %s [option] [-c <cpunumber>]\n"
	"Options:\n"
	"    -h,  --help           Print out this help and exit\n"
	"    -v,  --version        Print out version number and exit\n"
	"    -d,  --display  <dpy> The X11 display to connect to\n"
	"    -m,  --milisecs <ms>  The number of milisecs between updates\n"
	"    -b,  --bgcolor  <col> The background color\n"
	"    -o,  --oncolor  <col> The color for text and on leds\n"
	"    -f,  --offcolor <col> The color for off leds\n"
	"    -w,  --window         Run in a window\n"
	"    -nw, --no-window      Run as a dockapp\n"
	"    -s,  --shape          Use XShape extension\n"
	"    -ns, --no-shape       Don't use XShape extension\n"
	"    -c,  --cpunum   <cpu> The cpu to monitor (for SMP systems)\n",
	argv[0]);
 exit(0);
}

void print_version()
{
 printf(WINDOW_NAME " version " VERSION "\n");
 exit(0);
}

void parse_args()
{
 int n;

 for(n = 1; n < argc; n++)
 {
  if(!strcmp(argv[n], "-h") ||
     !strcmp(argv[n], "--help"))
  {
   print_usage();
  }
  else if(!strcmp(argv[n], "-v") ||
          !strcmp(argv[n], "--version"))
  {
   print_version();
  }
  else if(!strcmp(argv[n], "-d") ||
     !strcmp(argv[n], "--display"))
  {
   if(argc <= ++n)
   {
    bad_option(n);
   }
   opt_display = argv[n];
  }
  else if(!strcmp(argv[n], "-m") ||
          !strcmp(argv[n], "--milisecs"))
  {
   if(argc <= ++n)
   {
    bad_option(n);
   }
   opt_milisecs = strtol(argv[n], NULL, 0);
  }
  else if(!strcmp(argv[n], "-b") ||
          !strcmp(argv[n], "--bgcolor"))
  {
   if(argc <= ++n)
   {
    bad_option(n);
   }
   opt_bgcolor = argv[n];
  }
  else if(!strcmp(argv[n], "-o") ||
          !strcmp(argv[n], "--oncolor"))
  {
   if(argc <= ++n)
   {
    bad_option(n);
   }
   opt_oncolor = argv[n];
  }
  else if(!strcmp(argv[n], "-f") ||
          !strcmp(argv[n], "--offcolor"))
  {
   if(argc <= ++n)
   {
    bad_option(n);
   }
   opt_offcolor = argv[n];
  }
  else if(!strcmp(argv[n], "-w") ||
          !strcmp(argv[n], "--window"))
  {
   opt_window = 1;
  }
  else if(!strcmp(argv[n], "-nw") ||
          !strcmp(argv[n], "--no-window"))
  {
   opt_window = 0;
  }
  else if(!strcmp(argv[n], "-s") ||
          !strcmp(argv[n], "--shape"))
  {
   opt_shape = 1;
  }
  else if(!strcmp(argv[n], "-ns") ||
          !strcmp(argv[n], "--no-shape"))
  {
   opt_shape = 0;
  }
  else if(!strcmp(argv[n], "-c") ||
          !strcmp(argv[n], "--cpu"))
  {
   if(argc <= ++n)
   {
    bad_option(n);
   }
   opt_cpunum = strtol(argv[n], NULL, 0);
  }
  else
  {
   fprintf(stderr, "Bad option. Try `%s --help' for more information.\n", argv[0]);
   exit(1);
  }
 }
}

void make_window()
{
 XSetWindowAttributes  windowattributes;
 XGCValues             gcvalues;
 XpmAttributes         xpmattributes;
 XSizeHints           *sizehints;
 XWMHints             *wmhints;
 XClassHint           *classhint;
 XpmColorSymbol        xpmsymbols[3];
 GC                    tmpgc;
 int                   shapeevent, shapeerror; 

 display = XOpenDisplay(opt_display);
 if(!display)
 {
  printf("Cannot connect to display `%s'.\n", opt_display);
  exit(1);
 }
 screen = DefaultScreen(display);
 rootwindow = RootWindow(display, screen);
 colormap = DefaultColormap(display, screen);
 visual = DefaultVisual(display, screen);
 depth = DefaultDepth(display, screen);
 bgcolor = get_color(opt_bgcolor);
 oncolor = get_color(opt_oncolor);
 offcolor = get_color(opt_offcolor);
 gcvalues.graphics_exposures = False;
 gcvalues.function = GXcopy;
 gcvalues.foreground = oncolor;
 gcvalues.background = bgcolor;
 gc = XCreateGC(display, rootwindow,
                GCGraphicsExposures |
		GCFunction |
		GCForeground |
		GCBackground,
		&gcvalues);
 xpmattributes.valuemask = XpmColormap | XpmCloseness | XpmColorSymbols;
 xpmattributes.colormap = colormap;
 xpmattributes.closeness = 65536;
 xpmsymbols[0].name = "Background";
 xpmsymbols[0].value = opt_bgcolor;
 xpmsymbols[1].name = "On";
 xpmsymbols[1].value = opt_oncolor;
 xpmsymbols[2].name = "Off";
 xpmsymbols[2].value = opt_offcolor;
 xpmattributes.colorsymbols = xpmsymbols;
 xpmattributes.numsymbols = 3;
 XpmCreatePixmapFromData(display, rootwindow, pixmap_xpm,
                         &pixmap, &pixmask, &xpmattributes);
 XpmCreatePixmapFromData(display, rootwindow, icon_xpm,
                         &iconpixmap, None, NULL);
 buffer = XCreatePixmap(display, rootwindow,
                        WINDOW_WIDTH, WINDOW_HEIGHT, depth);
 XSetClipMask(display, gc, pixmask);
 XSetClipOrigin(display, gc, WINDOW_WIDTH, 0);
 XCopyArea(display, pixmap, pixmap, gc, 
           0, 0, WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_WIDTH, 0);
 XSetClipMask(display, gc, None);
 XSetClipOrigin(display, gc, 0, 0);
 XCopyArea(display, pixmap, pixmap, gc,
           WINDOW_WIDTH, 0, WINDOW_WIDTH, WINDOW_HEIGHT, 0, 0);
 XCopyArea(display, pixmap, buffer, gc,
           0, 0, WINDOW_WIDTH, WINDOW_HEIGHT, 0, 0);
 tmpgc = XCreateGC(display, pixmask, 0, NULL);
 XSetForeground(display, tmpgc, 0);
 XFillRectangle(display, pixmask, tmpgc,
                WINDOW_WIDTH, 0, WINDOW_WIDTH, WINDOW_HEIGHT);
 XFillRectangle(display, pixmask, tmpgc,
                0, WINDOW_HEIGHT, WINDOW_WIDTH * 2, WINDOW_HEIGHT);
 XFreeGC(display, tmpgc);
 windowattributes.background_pixmap = pixmap;
 windowattributes.colormap = colormap;
 windowattributes.event_mask = ExposureMask | 
                               KeyPressMask |
                               KeyReleaseMask |
			       ButtonPressMask |
			       ButtonReleaseMask |
			       PropertyChangeMask;
 window = XCreateWindow(display, rootwindow,
                        0, 0, WINDOW_WIDTH, WINDOW_HEIGHT,
			0, depth, InputOutput, visual,
			CWBackPixmap |
			CWColormap |
			CWEventMask,
			&windowattributes);
 iconwindow = XCreateWindow(display, rootwindow,
                            0, 0, WINDOW_WIDTH, WINDOW_HEIGHT,
			    0, depth, InputOutput, visual,
			    CWBackPixmap |
			    CWColormap |
			    CWEventMask,
			    &windowattributes);
 if(XShapeQueryExtension(display, &shapeevent, &shapeerror) && opt_shape)
 {
  XShapeCombineMask(display, window, ShapeBounding,
                    0, 0, pixmask, ShapeSet);
  XShapeCombineMask(display, iconwindow, ShapeBounding,
                    0, 0, pixmask, ShapeSet);
 }
 mapwindow = opt_window ? iconwindow : window;
 XStoreName(display, mapwindow, WINDOW_NAME);
 XSetIconName(display, mapwindow, ICON_NAME);
 sizehints = XAllocSizeHints();
 sizehints->flags = PMinSize | PMaxSize;
 sizehints->min_width = WINDOW_WIDTH;
 sizehints->min_height = WINDOW_HEIGHT;
 sizehints->max_width = WINDOW_WIDTH;
 sizehints->max_height = WINDOW_HEIGHT;
 XSetWMNormalHints(display, mapwindow, sizehints);
 XFree(sizehints);
 wmhints = XAllocWMHints();
 wmhints->flags = InputHint |
                  WindowGroupHint |
		  IconWindowHint |
		  IconPixmapHint |
		  StateHint;
 wmhints->input = True;
 wmhints->window_group = window;
 wmhints->icon_window = iconwindow;
 wmhints->icon_pixmap = iconpixmap;
 wmhints->initial_state = WithdrawnState;
 XSetWMHints(display, window, wmhints);
 XFree(wmhints);
 classhint = XAllocClassHint();
 classhint->res_name = CLASS_NAME;
 classhint->res_class = CLASS_CLASS;
 XSetClassHint(display, mapwindow, classhint);
 XFree(classhint);
 XSetCommand(display, window, argv, argc);
 wm_delete_window = XInternAtom(display, "WM_DELETE_WINDOW", False);
 XSetWMProtocols(display, mapwindow, &wm_delete_window, 1);
 if(opt_window)
 {
  mwmhints.flags = MWM_HINTS_DECORATIONS;
  mwmhints.decorations = 0;
  _motif_wm_hints = XInternAtom(display, "_MOTIF_WM_HINTS", 0);
  XChangeProperty(display, mapwindow,
                  _motif_wm_hints,
                  _motif_wm_hints,
                  32, PropModeReplace,
                  (unsigned char *)&mwmhints, 5);
 }
 XMapRaised(display, mapwindow);
}

void update_window()
{
 XCopyArea(display, buffer, window, gc,
           0, 0, WINDOW_WIDTH, WINDOW_WIDTH, 0, 0);
 XCopyArea(display, buffer, iconwindow, gc,
           0, 0, WINDOW_WIDTH, WINDOW_WIDTH, 0, 0);
 XSync(display, False);
}

void process_events()
{
 KeySym keysym;

 XNextEvent(display, &event);
 switch(event.type)
 {
  case Expose:
   update_window();
   break;
  case KeyPress:
   keysym = XLookupKeysym(&event.xkey, 0);
   switch(keysym)
   {
    case XK_Escape:
    case XK_Q:
    case XK_q:
     exitloop = 1;
     break;
   }
   break;
  case ClientMessage:
   if(event.xclient.data.l[0] == wm_delete_window)
    exitloop = 1;
   break;
 }
}

void draw_digit(int x, int y, int n)
{
 if(n>=0)
  XCopyArea(display, pixmap, buffer, gc,
            n * 8, 64, 8, 8, x, y);
 else
  XFillRectangle(display, buffer, gc, x, y, 8, 8);  
}

void draw_percent(int x, int y, int n)
{
 if(n == 100)
 draw_digit(x, y, 1);
 else draw_digit(x, y, -1);
 if(n > 9) draw_digit(x + 8, y, (n % 100) / 10);
 else draw_digit(x + 8, y, -1);
 draw_digit(x + 16, y, (n % 10));
}

void draw_graph(int x, int y, int mark_w, int mark_h, int height, int value)
{
 int n;

 for(n = 0; n < height; n++)
 {
  XSetForeground(display, gc,
                 (value >= height - n) ? oncolor : offcolor);
  XFillRectangle(display, buffer, gc, 
                 x, y + n * mark_h,
                 mark_w - 1, mark_h - 1);
 }
}

void draw_window()
{
 int y;

 XSetForeground(display, gc, bgcolor);
 draw_digit(CPUNUM_X, CPUNUM_Y, opt_cpunum);
 draw_percent(PERCENT_X, PERCENT_Y, cpu_used);
 XCopyArea(display, buffer, buffer, gc,
           GRAPH_X+GRAPH_PT_W, GRAPH_Y,
	   (GRAPH_LENGTH - 1) * GRAPH_PT_W,
	   GRAPH_HEIGHT * GRAPH_PT_H,
	   GRAPH_X, GRAPH_Y);
 for(y = 0; y < GRAPH_HEIGHT; y++)
 {
  draw_graph(GRAPH_X + (GRAPH_LENGTH - 1) * GRAPH_PT_W,
             GRAPH_Y, GRAPH_PT_W, GRAPH_PT_H,
	     GRAPH_HEIGHT,
	     cpu_used * GRAPH_HEIGHT / 100);
 }
}

void handle_timer(int sig)
{
 if(sig == SIGALRM)
 {
  cpu_getusage();
  draw_window();
  update_window();
 }
}

void start_timer()
{
 struct itimerval itv;

 signal(SIGALRM, handle_timer);
 itv.it_value.tv_sec = 1;
 itv.it_value.tv_usec = 0;
 itv.it_interval.tv_sec = opt_milisecs / 1000;
 itv.it_interval.tv_usec = (opt_milisecs % 1000) * 1000;
 setitimer(ITIMER_REAL, &itv, NULL);
}

void stop_timer()
{
 struct itimerval itv;

 signal(SIGALRM, SIG_IGN);
 itv.it_value.tv_sec = 0;
 itv.it_value.tv_usec = 0;
 itv.it_interval.tv_sec = 0;
 itv.it_interval.tv_usec = 0;
 setitimer(ITIMER_REAL, &itv, NULL);
}

void free_stuff()
{
 XUnmapWindow(display, mapwindow);
 XDestroyWindow(display, iconwindow);
 XDestroyWindow(display, window);
 XFreePixmap(display, buffer);
 XFreePixmap(display, iconpixmap);
 XFreePixmap(display, pixmask);
 XFreePixmap(display, pixmap);
 XFreeGC(display, gc);
 XCloseDisplay(display);
}

int main(int carg, char **varg)
{
 argc = carg;
 argv = varg;
 parse_args();
 make_window();
 start_timer();
 cpu_init(opt_cpunum);
 while(!exitloop)
 {
  process_events();
 }
 stop_timer();
 free_stuff();
 return 0;
}
